import xbmcvfs

from resources.lib.const import COMMAND, LANG, DOWNLOAD_STATUS, SERVICE, DOWNLOAD_SERVICE_EVENT
from resources.lib.gui import InfoDialog
from resources.lib.gui.renderers.dialog_renderer import DialogRenderer
from resources.lib.storage.sqlite import DB
from resources.lib.utils.addtolib import add_to_library, update_tv_shows
from resources.lib.utils.download import get_file_path
from resources.lib.utils.kodiutils import show_settings, get_string, get_log_file_path, read_log, clean_log, post_log, \
    show_log_dialog, refresh, show_settings_with_focus
from resources.lib.api.trakt_api import trakt
from resources.lib.utils.speedtest import speed_test


class Command:
    def __init__(self, plugin_core):
        self._plugin_core = plugin_core

        self.command_map = {
            COMMAND.OPEN_SETTINGS: self.show_settings,
            COMMAND.SHOW_SETTINGS_WITH_FOCUS: self.show_settings_with_focus,
            COMMAND.SHOW_LOG_PATH: self.show_log_path,
            COMMAND.UPLOAD_LOG: self.upload_log,
            COMMAND.SET_PROVIDER_CREDENTIALS: self.set_provider_credentials,
            COMMAND.SET_SUBTITLES_CREDENTIALS: DialogRenderer.set_subtitles_credentials,
            COMMAND.REFRESH_PROVIDER_TOKEN: self._plugin_core.auth.refresh_provider_token,
            COMMAND.ADD_TO_LIBRARY: add_to_library,
            COMMAND.DELETE_HISTORY_ALL: self.delete_history_all,
            COMMAND.PIN: self.pin,
            COMMAND.UNPIN: self.unpin,
            COMMAND.HIDE_MENU_ITEM: self.hide_menu_item,
            COMMAND.SHOW_MENU_ITEM: self.show_menu_item,
            COMMAND.DELETE_HISTORY_ITEM: self.delete_history_item,
            COMMAND.DELETE_DOWNLOAD_ITEM: self.delete_download_item,
            COMMAND.CANCEL_DOWNLOAD_ITEM: self.cancel_download_item,
            COMMAND.RESUME_DOWNLOAD_ITEM: self.resume_download_item,
            COMMAND.PAUSE_DOWNLOAD_ITEM: self.pause_download_item,
            COMMAND.UPDATE_TV_SHOWS: update_tv_shows,
            COMMAND.SPEED_TEST: speed_test,
        }

        if trakt.is_authenticated():
            self.command_map.update({
                COMMAND.TRAKT_LIST_APPEND: trakt.list_append,
                COMMAND.TRAKT_LIST_CLONE: trakt.list_clone,
                COMMAND.TRAKT_LIST_DELETE: trakt.list_delete,  # trakt_renderer?
                COMMAND.TRAKT_LIST_LIKE: trakt.list_like,
                COMMAND.TRAKT_LIST_RENAME: trakt.list_rename,
                COMMAND.TRAKT_LIST_UNLIKE: trakt.list_unlike,
                COMMAND.TRAKT_UNFOLLOW: trakt.unfollow,
                COMMAND.TRAKT_LOGOUT: trakt.logout,
                COMMAND.TRAKT_MEDIA_MENU: trakt.media_menu,
            })
        else:
            self.command_map[COMMAND.TRAKT_LOGIN] = trakt.login

    def __call__(self, command, *args, **kwargs):
        return self.command_map[command]

    @staticmethod
    def show_settings():
        show_settings()

    @staticmethod
    def show_settings_with_focus(id1, id2):
        show_settings_with_focus(int(id1), int(id2))

    @staticmethod
    def pin(dir_route, key):
        DB.PINNED.add(dir_route, key)
        refresh()

    @staticmethod
    def unpin(dir_route, key):
        DB.PINNED.delete(dir_route, key)
        refresh()

    @staticmethod
    def hide_menu_item(dir_route, key):
        DB.HIDDEN.add(dir_route, key)
        refresh()

    @staticmethod
    def show_menu_item(dir_route, key):
        DB.HIDDEN.delete(dir_route, key)
        refresh()

    @staticmethod
    def show_log_path():
        DialogRenderer.ok(get_string(LANG.LOG_FILE_PATH), get_log_file_path())

    @staticmethod
    def upload_log():
        success, data = read_log()
        if success:
            content = clean_log(data)
            success, response = post_log(content)
            if success:
                show_log_dialog(response)
            else:
                DialogRenderer.ok(get_string(LANG.LOG_FILE_PATH), 'Failed to upload log file: ' + response)
        else:
            DialogRenderer.ok(get_string(LANG.LOG_FILE_PATH), 'Failed to get log file: ' + data)

    def set_provider_credentials(self):
        self._plugin_core.auth.set_provider_credentials()

    @staticmethod
    def delete_history_item(media_id):
        DB.WATCH_HISTORY.delete(media_id=media_id)
        refresh()

    @staticmethod
    def delete_history_all():
        DB.WATCH_HISTORY.delete()
        InfoDialog(get_string(LANG.HISTORY_DELETED)).notify()

    @staticmethod
    def delete_download_item(dl_id):
        item = DB.DOWNLOAD.get(dl_id)
        DB.DOWNLOAD.delete(dl_id=dl_id)
        xbmcvfs.delete(get_file_path(item[1], item[3]))
        refresh()

    @staticmethod
    def cancel_download_item(dl_id):
        item = DB.DOWNLOAD.get(dl_id)
        if item[4] == DOWNLOAD_STATUS.PAUSED:
            xbmcvfs.delete(get_file_path(item[1], item[3]))
        DB.DOWNLOAD.cancel_download(dl_id=dl_id, status=DOWNLOAD_STATUS.CANCELLED)
        refresh()

    def resume_download_item(self, dl_id):
        DB.DOWNLOAD.set_status(dl_id, DOWNLOAD_STATUS.QUEUED)
        self._plugin_core.send_service_message(SERVICE.DOWNLOAD_SERVICE, DOWNLOAD_SERVICE_EVENT.ITEM_RESUMED,
                                               download_id=dl_id)
        refresh()

    @staticmethod
    def pause_download_item(dl_id):
        DB.DOWNLOAD.set_status(dl_id, DOWNLOAD_STATUS.PAUSED)
        refresh()
